#include <std_include.hpp>
#include "loader/component_loader.hpp"
#include "game/game.hpp"
#include "game/dvars.hpp"

#include "command.hpp"
#include "console.hpp"
#include "fastfiles.hpp"


#include <utils/hook.hpp>
#include <utils/memory.hpp>
#include <utils/string.hpp>
#include <utils/io.hpp>

namespace fastfiles
{
	namespace
	{
		utils::hook::detour db_try_load_x_file_internal_hook;
		utils::hook::detour db_find_x_asset_header_hook;

		template <typename T> inline void merge(std::vector<T>* target, T* source, size_t length)
		{
			if (source)
			{
				for (size_t i = 0; i < length; ++i)
				{
					target->push_back(source[i]);
				}
			}
		}

		template <typename T> inline void merge(std::vector<T>* target, std::vector<T> source)
		{
			for (auto& entry : source)
			{
				target->push_back(entry);
			}
		}

		void db_try_load_x_file_internal(const char* zone_name, const int zone_flags, const int is_base_map)
		{
			console::info("Loading fastfile %s\n", zone_name);
			db_try_load_x_file_internal_hook.invoke<void>(zone_name, zone_flags, is_base_map);
		}

		void dump_gsc_script(const std::string& name, game::XAssetHeader header)
		{
			if (!dvars::g_dump_scripts->current.enabled)
			{
				return;
			}

			std::string buffer;
			buffer.append(header.scriptfile->name, std::strlen(header.scriptfile->name) + 1);
			buffer.append(reinterpret_cast<char*>(&header.scriptfile->compressedLen), 4);
			buffer.append(reinterpret_cast<char*>(&header.scriptfile->len), 4);
			buffer.append(reinterpret_cast<char*>(&header.scriptfile->bytecodeLen), 4);
			buffer.append(header.scriptfile->buffer, header.scriptfile->compressedLen);
			buffer.append(reinterpret_cast<char*>(header.scriptfile->bytecode), header.scriptfile->bytecodeLen);

			const auto out_name = std::format("gsc_dump/{}.gscbin", name);
			utils::io::write_file(out_name, buffer);

			console::info("Dumped %s\n", out_name.data());
		}


		game::XAssetHeader db_find_x_asset_header_stub(game::XAssetType type, const char* name, int allow_create_default)
		{
			const auto start = game::Sys_Milliseconds();
			const auto result = db_find_x_asset_header_hook.invoke<game::XAssetHeader>(type, name, allow_create_default);
			const auto diff = game::Sys_Milliseconds() - start;

			if (type == game::ASSET_TYPE_SCRIPTFILE)
			{
				dump_gsc_script(name, result);
			}

			if (diff > 100)
			{
				console::print(
					result.data == nullptr
						? console::con_type_error
						: console::con_type_warning,
					"Waited %i msec for asset '%s' of type '%s'.\n",
					diff,
					name,
					game::g_assetNames[type]
				);
			}

			return result;
		}

		void reallocate_asset_pool(const game::XAssetType type, const unsigned int new_size)
		{
			const size_t element_size = game::DB_GetXAssetTypeSize(type);

			auto* new_pool = utils::memory::get_allocator()->allocate(new_size * element_size);
			std::memmove(new_pool, game::DB_XAssetPool[type], game::g_poolSize[type] * element_size);

			game::DB_XAssetPool[type] = new_pool;
			game::g_poolSize[type] = new_size;
		}

		void p_mem_free_stub(const char* name, game::PMem_Direction alloc_dir)
		{
			console::info("Unloaded fastfile %s\n", name);
			game::PMem_Free(name, alloc_dir);
		}

		const auto skip_extra_zones_stub = utils::hook::assemble([](utils::hook::assembler& a)
		{
			const auto skip = a.newLabel();
			const auto original = a.newLabel();

			a.pushad64();
			a.test(ebp, game::DB_ZONE_CUSTOM); // allocFlags
			a.jnz(skip);

			a.bind(original);
			a.popad64();
			a.mov(rdx, 0x140835F28);
			a.mov(rcx, rsi);
			a.call_aligned(strcmp);
			a.jmp(0x1403217C0);

			a.bind(skip);
			a.popad64();
			a.mov(r15d, 0x80);
			a.not_(r15d);
			a.and_(ebp, r15d);
			a.jmp(0x1403217F6);
		});
	}

	bool exists(const std::string& zone_name)
	{
		auto is_localized = game::DB_IsLocalized(zone_name.data());
		auto handle = game::Sys_CreateFile(game::Sys_Folder(is_localized), utils::string::va("%s.ff", zone_name.data()));
		if (handle != (HANDLE)-1)
		{
			CloseHandle(handle);
			return true;
		}
		return false;
	}

	void enum_assets(const game::XAssetType type, const std::function<void(game::XAssetHeader)>& callback, const bool include_override)
	{
		game::DB_EnumXAssets_Internal(type, static_cast<void(*)(game::XAssetHeader, void*)>([](game::XAssetHeader header, void* data)
		{
			const auto& cb = *static_cast<const std::function<void(game::XAssetHeader)>*>(data);
			cb(header);
		}), &callback, include_override);
	}

	void Load_CommonZones(game::XZoneInfo* zoneInfo, unsigned int zoneCount, game::DBSyncMode syncMode)
	{
		std::vector<game::XZoneInfo> data;
		merge(&data, zoneInfo, zoneCount);

		if (fastfiles::exists("mod"))
		{
			data.push_back({ "mod", game::DB_ZONE_COMMON | game::DB_ZONE_CUSTOM, 0 });
		}

		game::DB_LoadXAssets(data.data(), static_cast<std::uint32_t>(data.size()), syncMode);
	}

	void Load_LevelZones(game::XZoneInfo* zoneInfo, unsigned int zoneCount, game::DBSyncMode syncMode)
	{
		std::vector<game::XZoneInfo> data;
		merge(&data, zoneInfo, zoneCount);

		if (fastfiles::exists("mod"))
		{
			data.push_back({ "mod", game::DB_ZONE_COMMON | game::DB_ZONE_CUSTOM, 0 });
		}

		game::DB_LoadXAssets(data.data(), static_cast<std::uint32_t>(data.size()), syncMode);
	}

	utils::hook::detour sys_createfile_hook;
	HANDLE sys_create_file_stub(game::Sys_Folder folder, const char* base_filename)
	{
		auto* fs_basepath = game::Dvar_FindVar("fs_basepath");
		auto* fs_game = game::Dvar_FindVar("fs_game");

		std::string dir = fs_basepath ? fs_basepath->current.string : "";
		std::string mod_dir = fs_game ? fs_game->current.string : "";

		if (base_filename == "mod.ff"s)
		{
			if (!mod_dir.empty())
			{
				auto path = utils::string::va("%s\\%s\\%s", dir.data(), mod_dir.data(), base_filename);
				if (utils::io::file_exists(path))
				{
					return CreateFileA(path, 0x80000000, 1u, 0, 3u, 0x60000000u, 0);
				}
			}
			return (HANDLE)-1;
		}

		return sys_createfile_hook.invoke<HANDLE>(folder, base_filename);
	}

	class component final : public component_interface
	{
	public:
		void post_unpack() override
		{
			db_try_load_x_file_internal_hook.create(SELECT_VALUE(0x140275850, 0x1403237F0), &db_try_load_x_file_internal);

			db_find_x_asset_header_hook.create(game::DB_FindXAssetHeader, db_find_x_asset_header_stub);
			dvars::g_dump_scripts = game::Dvar_RegisterBool("g_dumpScripts", false, game::DVAR_FLAG_NONE, "Dump GSC scripts to binary format");

			utils::hook::call(SELECT_VALUE(0x1402752DF, 0x140156350), p_mem_free_stub);
			utils::hook::call(SELECT_VALUE(0x140276004, 0x140324259), p_mem_free_stub);

			// Allow loading of unsigned fastfiles
			utils::hook::set<uint8_t>(0x1402FBF23, 0xEB); // DB_LoadXFile
			utils::hook::nop(0x1402FC445, 2); // DB_SetFileLoadCompressor

			// Don't load eng_ + patch_ with loadzone
			utils::hook::nop(0x1403217B1, 15);
			utils::hook::jump(0x1403217B1, skip_extra_zones_stub, true);

			// Add custom zone paths
			sys_createfile_hook.create(game::Sys_CreateFile, sys_create_file_stub);

			// Load our custom fastfiles (Mod)
			utils::hook::call(0x1405E7113, Load_CommonZones);

			// Reload mod after level is loaded to allow overriding map stuff
			utils::hook::call(0x140320ED1, Load_LevelZones);

			command::add("materiallist", [](const command::params& params)
			{
				game::DB_EnumXAssets_FastFile(game::ASSET_TYPE_MATERIAL, [](const game::XAssetHeader header, void*)
				{
					if(header.material && header.material->name)
					{
						console::info("%s\n", header.material->name);
					}
				}, nullptr, false);
			});

			if (!game::environment::is_sp())
			{
				reallocate_asset_pool(game::ASSET_TYPE_WEAPON, 320);
			}
		}
	};
}

REGISTER_COMPONENT(fastfiles::component)
